from collections import deque
import importlib
import itertools
from shlex import shlex
import sys

from attrs import Attrs
from log import log
from rule import Rule


class ChartParser:
    def __init__(self, rules=[], rule_files=[], argv=[]):
        self.options = dict(
            # original only, no dependencies yet (--dependency/-d)
            print_vertical=False,
            # zatím jednoduchý
            print_graph=False,
            # stop after each processed sentence
            interactive=False,
            )
        if rules:
            self.load_rules()
        else:
            if not rule_files and argv:
                rule_files, options = self.read_cmdline(argv[1:])
                self.options.update(options)

            if not rule_files:
                rule_files = ['cz']
            self.load_rules_from_files(rule_files)

        self.print_rules()

    def parse(self, tokens, root='S'):
        self.tokens = tokens
        sentence = None
        closing = None

        if tokens[0].get('structural') in ('s', 'head'):
            sentence = tokens.pop(0)
            log.info(sentence)

        if tokens[-1].get('structural') in ('s', 'head'):
            closing = tokens.pop()

        self._add_parsed_attibutes_to_tokens()

        for token in tokens:
            log.info(Attrs(token))

        if closing:
            log.info(closing)

        # když mezi nima není sloveso nebo „lze“ k6, spustit parse(root='NP')
        if False:
            log.info('No verb found, parsing as NP instead')

        self.chart = []  # jak by to mohla bejt množina, když pro Edge neumím
                         # udělat __hash__? navíc jsou její podstatné části,
                         # dokonce už left, slovníky
        self.root = root
        self.prepare_agenda()

        while self.agenda:
            edge = self.agenda.popleft()
            # it might be tempting to add the edge to the chart right away
            # to simplify edge_is_duplicated(); but keep in mind that the edge
            # would “want to” combine with itself in propose_edges()
            log.info('             %s', edge)
            for new_edge, message in self.propose_edges(edge):
                if not self.edge_is_duplicated(new_edge, edge):
                    log.info(message)
                    self.agenda.append(new_edge)
                else:
                    log.warning(message)
            self.chart.append(edge)

        self.print_chart()
        length = len(self.tokens)
        self.successful_edges = [edge for edge in self.chart if edge.complete
                                 and edge.begin == 0 and edge.end == length]
        if not self.successful_edges:
            log.error('\nNo edge covering the whole input!\n')
        for edge in self.successful_edges:
            log.info('\nPassed!      %s\n', edge)
            if self.options['print_graph']:
                self.print_graph(edge)
        if self.options['print_vertical']:
            if sentence:
                sentence['passed'] = 'yes' if self.successful_edges else 'no'
                print(sentence)
            for token in tokens:
                print(token.vertical())
            if closing:
                print(closing)
        # if self.interactive:
        #     input()

    def edge_is_duplicated(self, edge, agenda_edge):
        return edge == agenda_edge or edge in self.agenda or edge in self.chart

    def print_chart(self):
        for edge in self.chart:
            # vypisuju i částečně uplatněné hrany a dokonce hrany s ε-pravidly
            # TODO: rozdělit výpis na hotové hrany (včetně ε-pravidel, ale tam
            # přeskakovat ty s příznakem „hidden“) a částečné (ale z těch asi
            # vybrat ty, které jsou zahrnuté v uzavřených hranách)
            # if edge.point == 0 and edge.rule.right:
            if edge.complete and edge.right:
                log.info('chart        %s', edge)

    def print_rules(self):
        for rule in self.rules:
            log.debug('rule         %s', rule)
        log.info('\n')

    @staticmethod
    def extract_attributes_from_tag(tag):
        if ',' in tag:
            tags = tag.split(',')
            if '' not in tags:
                return ChartParser.merge_ambiguous_tags(tags)
        elif len(tag) % 2 != 0:
            raise Exception('Malformed tag "%s"', tag)
        return ((attr, val) for (attr, val) in grouper(tag, 2))

    @staticmethod
    def merge_ambiguous_tags(tags):
        merged = {}
        for attr, val in itertools.chain.from_iterable(
                ChartParser.extract_attributes_from_tag(tag) for tag in tags):
            if attr not in merged:
                merged[attr] = val
            elif isinstance(merged[attr], str):
                if val != merged[attr]:
                    merged[attr] = [merged[attr], val]
            elif val not in merged[attr]:
                merged[attr].append(val)
        return merged

    def _add_parsed_attibutes_to_tokens(self):
        for token in self.tokens:
            if 'structural' in token:  # to možná nebude potřeba; nemělo by
                continue
            if 'tag' in token:
                token.update(
                    self.extract_attributes_from_tag(token['tag']))

    def parse_from_vertical(self, lines=sys.stdin):
        tokens = []
        phrase = None
        for line in lines:
            line = line.strip()
            if line.startswith('<'):
                tag, closing, empty, attrs = self.parse_structural(line)
                attrs = dict(attrs)
                structural = (Attrs(structural=tag, closing=closing,
                                    empty=empty, line=line, **attrs))
                if tag in ('s', 'head'):
                    if closing or tokens:
                        tokens.append(structural)
                        self.parse(tokens)
                        tokens = []
                    if not closing:
                        tokens.append(structural)
                    continue

                if self.options['print_vertical']:
                    print(structural)

                if tag == 'g':  # and empty
                    if tokens:
                        tokens[-1]['no-space'] = True  # tokens glued together
                elif tag == 'phr' and not closing:  # MWE
                    phrase = []
                    phr_lemma = attrs.get('l')
                    phr_tag = attrs.get('t')
                    # TODO: opravit práci se shlex!!
                    if phr_tag.startswith('"') and phr_tag.startswith('"'):
                        phr_tag = phr_tag[1:-1]
                elif tag == 'phr' and closing:
                    token = Attrs(word=' '.join(phrase), lemma=phr_lemma,
                                  tag=phr_tag, mwe=True)
                    tokens.append(token)
                    phrase = None
                elif tag in ('doc', 'p'):
                    pass  # zatím neřeším (parseru to může být skoro jedno)
                elif tag == 'table':
                    table = True  # přeskakovat někdy celý obsah?
                elif tag == 'sign':
                    pass  # zajímavé: Josef Juřeník, přednosta stanice
            elif line:
                try:
                    word, lemma, tag = line.split()
                except ValueError:
                    log.warning('unexpected line ' + line)
                    word = lemma = line
                    if line in ',.():':
                        tag = 'kI'  # cesta do pekel :-D
                    elif line[0] in '0123456789':  # TODO: vylepšit heuristiku
                        tag = 'k4'
                    else:  # 31. března
                        tag = 'k?'
                if phrase is not None:
                    phrase.append(line)
                else:
                    token = Attrs(word=word, lemma=lemma, tag=tag, line=line)
                    tokens.append(token)

        if tokens:
            self.parse(tokens)

    @staticmethod
    def parse_structural(line):
        parser = shlex(line[1:])
        attrs = []
        closing = None
        tag = None
        empty = None
        attr = None
        value = None
        while True:
            token = parser.get_token()
            if not token:
                break
            elif closing is None:
                closing = token == '/'
                if not closing:
                    tag = token
            elif tag is None:
                if token == '>':
                    break
                tag = token
            elif attr is None:
                if token == '/':
                    empty = True
                elif token == '>':
                    break
                elif token == '=':
                    break  # unexpected
                else:
                    attr = token
                    value = '='
            elif empty:
                if token == '>':
                    break
                else:
                    break  # unexpected
            elif value == '=':
                if token == '=':
                    value = None
                elif token == '/':
                    attrs.append((attr, None))
                    empty = True
                elif token == '>':
                    attrs.append((attr, None))
                    break
                else:
                    attrs.append((attr, None))
                    attr = token
            elif value is None:
                attrs.append((attr, token))
                attr = None
                value = None
        return tag, closing, empty, attrs

    def load_rules_from_files(self, rule_files):
        self.rules = []
        for rules_filename in rule_files:
            rules_file = importlib.import_module('rules.' + rules_filename)
            self.load_rules(rules_file.rules)

    def load_rules(self, rules):
        for rule in rules:
            if not isinstance(rule, Rule):
                rule = Rule(rule)
            self.rules.append(rule)
            for epsilon_rule in rule.epsilon_rules:
                self.rules.append(epsilon_rule)

    def print_graph(self, edge, duplicates=[], level=0):
        if edge is None or edge in duplicates:
            return

        print('  '*level, edge)
        level += 1
        duplicates.append(edge)  # TODO: tohle bude doufám jen obezlička!
        for child in edge.children:
            self.print_graph(child, duplicates, level)

    @staticmethod
    def read_cmdline(argv):
        rules = []
        options = dict(print_vertical=False, interactive=False)
        expect_rules = False
        for arg in argv:
            if arg == '--rules':
                expect_rules = True
                continue
            if expect_rules:
                rules.append(arg)
                expect_rules = False
            elif arg == '--print-vertical':
                options['print_vertical'] = True
            elif arg == '--print-graph':
                options['print_graph'] = True
            elif arg == '--interactive':
                options['interactive'] = True
            else:
                log.warning('Unknown parameter: ' + arg)
        return rules, options


# copied from oVirt/VDSM, which in turn used
# http://docs.python.org/2.6/library/itertools.html?highlight=grouper#recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.zip_longest(fillvalue=fillvalue, *args)
