from collections import defaultdict
from shlex import shlex

from attrs import Attrs, escape


class Rule:
    """
    Should be declared and kept as immutable so it can be hashed
    (including its subclass, Edge)
    """
    def __init__(self, rule=None, left=None, right=None):
        if rule is not None:
            try:
                left_side, right_side = rule.split(' → ')
            except ValueError:
                raise ValueError('Multiple rules stuck together? ' + rule)
            self.left = next(self.read_rule(left_side))
            self.right = list(self.read_rule(right_side))

        if left is not None:
            self.left = left  # update?
        if right is not None:
            self.right = right

        self._parse_point()
        self.epsilon_rules = list(self._process_optional_tokens())
        self.attr_matches = self._process_matches()

    def __str__(self):
        right = [str(attrs) for attrs in self.right]
        if 'point' in self.__dict__:
            right.insert(self.point, '•')
            if len(right) == 1:  # ε-rule with • now inserted
                right.insert(0, 'ε')  # or just right = list('ε•')
        elif not right:
            right = ['ε']
        span = ('{}–{} '.format(self.begin, self.end)
                if 'begin' in self.__dict__ else '')
        return span + str(self.left) + ' → ' + ' '.join(right)

    def __repr__(self):
        return 'Rule({})'.format(escape(self))

    def _process_matches(self):
        attr_matches = defaultdict(list)  # → OrderedDict, abych nemusel řadit
        for index, token in enumerate([self.left] + list(self.right), -1):
            for attr, value in token.items():
                if value is None:
                    attr_matches[attr].append(index)

        for attr, match_on_indexes in attr_matches.items():
            if len(match_on_indexes) == 1:
                raise ValueError('No other rule token to match "%s" on in %s'
                                 % (attr, self))
        return attr_matches

    @staticmethod
    def read_rule(rule):
        """
        Return a list of symbols, i.e. (non-)terminals from a string
        (which has already been split to the left/right side)
        """
        if rule.strip() == 'ε':
            return

        attrs = None  # not None ⇔ inside the token
        head = False  # expecting the phrase head (or nothing)
        attr = ''  # a = plus an attribute value may follow
        equal_sign = False  # we are expecting a value for an attribute after =
        closed = False  # the token has already ended with )

        for token in shlex(rule, posix=True):
            if token == '(' or (token == '•' and (attrs is None or closed)):
                if attrs is not None:
                    yield attrs
                    if not closed:  # TODO: allow ambiguous (a=( b=1) ?
                        raise ValueError('Unexpected "("')

                if token == '•':
                    yield '•'
                    attrs = None
                    continue

                attrs = Attrs()
                head = closed = False
            elif token == '?':
                attrs['optional'] = True
            elif token == ')':
                if equal_sign:
                    raise ValueError('Unterminated attribute ' + attr)
                elif attr:
                    attrs[attr] = None  # není lepší pro shodové atributy True?

                attr = ''
                equal_sign = False
                closed = True
            elif closed:
                raise ValueError('Outside token: ' + token)
            elif not attr:
                if token == '=':
                    raise ValueError('Unexpected "="')
                elif not head and token[0].isupper():
                    head = attrs['phrase'] = token
                    continue
                attr = token
            elif token == '=':
                equal_sign = True
            elif equal_sign:
                attrs[attr] = token
                attr = ''
                equal_sign = False
            else:
                attrs[attr] = None
                attr = token

        if attrs is not None:
            yield attrs
            if not closed:
                raise ValueError('Rule not terminated by ")"')

    def _process_optional_tokens(self):
        for index, token in enumerate(self.right, 1):
            if token.get('optional', False):
                if 'phrase' not in token:
                    new_token = Attrs(token)
                    del new_token['optional']

                    # token['word'] = None
                    token.pop('word', None)
                    token.pop('lemma', None)
                    token.pop('tag', None)
                    token['phrase'] = self.left['phrase'] + str(index)

                    new_head = Attrs(token)
                    del new_head['optional']
                    yield Rule(left=new_head, right=[new_token])
                yield Rule(left=Attrs(phrase=token['phrase']), right=[])

    def _parse_point(self):
        for index, token in enumerate(list(self.right)):
            if token == '•':
                if 'point' in self.__dict__:
                    raise ValueError('Duplicated • in edge ' + str(self))
                self.right.pop(index)
                self.point = index

    def insert_token(self, inserted_token, before=False, inside=True,
                     after=False, position=None):
        if isinstance(inserted_token, str):
            inserted_token = next(self.read_rule(inserted_token))

        if position is not None:
            new_right = [Attrs(token) for token in self.right]
            new_right.insert(position, Attrs(inserted_token))
            return Rule(left=Attrs(self.left), right=new_right)

        new_right = [Attrs(inserted_token)] if before else []
        if inside:
            right = new_right
            for token in self.right:
                right.append(Attrs(token))
                right.append(Attrs(inserted_token))
            if not after:
                right.pop()
            new_right = right
        else:
            new_right.extend(self.right)
            if after:
                new_right.append(Attrs(inserted_token))

        return Rule(left=Attrs(self.left), right=new_right)
